!
! 
!
program main
  !
  use precision
  use matrice_s, only: set_ref, s_mat, allocation_s, deallocation_s, init_invs
  use form_factor_type, only: form_factor, operator(-), operator(+), operator(*), assignment(=)
  use form_factor_2p ! export all
  use form_factor_3p ! export all
  use form_factor_4p ! export all
  use cache, only: allocate_cache, reset_cache, clear_cache
  use constante, only: s_null, pi
  use spinor, only: scalar  ! to get the function scalar
  use parametre, only: rat_or_tot_par
  !
  implicit none
  !
  type(form_factor) :: g1PPPP,g1PPPM,g1PPMM,g2PPPP,g2PPPM,g2PPMM,g3PPPP,g3PPPM,g3PPMM
  type(form_factor) :: ampPPPP,ampPPPM,ampPPMM
  type(form_factor), dimension(106) :: tc
  real(ki), dimension(4) :: k1,k2,k3,k4
  real(ki) :: ener,theta,t1,t2,S,T,U,resPPMM
  !
  !
  write (*,*) 'Evaluating 4 photon amplitude....'
  !
  ! Opening of the error files
  !
  open(unit=19,file='error_4photon.txt',status='unknown')
  write (19,*) 'errors'
  !
  ! Opening of the files containing the results
  !
  open(unit=17,file='result_4photon.txt',status='unknown')
  !
  ! rat_or_tot_par can have the value 'tot' or 'rat'
  ! overwrite the default value "tot" defined in the module parametre
  ! change 'tot' in 'rat'
  !
  rat_or_tot_par = 'tot'
  !
  ener   =  1.0_ki/2.0_ki
  theta  = 15.0_ki/180.0_ki*pi
  !
  k1 = (/ener,0.0_ki,0.0_ki,ener/)
  k2 = (/ener,0.0_ki,0.0_ki,-ener/)
  k3 =-(/ener,0.0_ki,ener*sin(theta),ener*cos(theta)/)
  k4 =-(/ener,0.0_ki,-ener*sin(theta),-ener*cos(theta)/) 
  !
  ! Mandelstam variables:
  ! 
  S = 2.0_ki*scalar(k1,k2)
  T = 2.0_ki*scalar(k2,k3)
  U = 2.0_ki*scalar(k1,k3)
  !
  write (17,*) '>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<'
  write (17,*) '>>>>>>>>>>>>> GOLEM OUTPUT <<<<<<<<<<<<<<'
  write (17,*) '>>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<'
  write (17,*) ''
  write (17,*) '  Four photon helicity amplitudes'
  write (17,*) '  ==============================='
  write (17,*) ''
  write (17,*) '  In all numbers alpha^2 times a phase is omitted !'
  write (17,*) ''
  write (17,*) '  phase(++++) = [21]*[43]/<12>/<34>'
  write (17,*) '  phase(+++-) = [21]*<14>*[31]/<12>/[41]/<13>'
  write (17,*) '  phase(++--) = [21]*<34>/<12>/[43]'
  write (17,*) ''
  write(17,'("   S = ",e16.10)') S
  write(17,'("   T = ",e16.10)') T
  write(17,'("   U = ",e16.10)') U
  !
  call allocation_s(4)
  !
  ! Set of initial propagators
  !
  set_ref = (/1,2,3,4/)  
  !
  call allocate_cache(4)
  !
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  ! GRAPH 1: 
  !
  !   k4->            <-k3 
  !   yyyyx----<----xyyyyy
  !       |         |
  !       |         |
  !       |         |
  !       |         |
  !   yyyyx---->----xyyyyy
  !   k2->            <-k1 
  ! 
  ! graph with reversed charge flow amounts to factor 2
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  !
  !
  ! Definition of the S matrix
  !
  s_mat(1,1) = 0.0_ki
  s_mat(1,2) = 0.0_ki
  s_mat(1,3) = S
  s_mat(1,4) = 0.0_ki
  !
  s_mat(2,1) = 0.0_ki
  s_mat(2,2) = 0.0_ki
  s_mat(2,3) = 0.0_ki
  s_mat(2,4) = U
  !
  s_mat(3,1) = s_mat(1,3)
  s_mat(3,2) = s_mat(2,3)
  s_mat(3,3) = 0.0_ki
  s_mat(3,4) = 0.0_ki
  !
  s_mat(4,1) = s_mat(1,4)
  s_mat(4,2) = s_mat(2,4)
  s_mat(4,3) = s_mat(3,4)
  s_mat(4,4) = 0.0_ki
  !
  call cpu_time(t1)
  !
  ! The inverse of the S matrix is computed numerically, and also all
  ! the related quantities: b coefficients, ......
  !
  call init_invs()
  !
  tc(1)=a40(s_null)
  tc(2)=a41(1,s_null)
  tc(3)=a41(2,s_null)
  tc(4)=a41(3,s_null)
  tc(5)=b42(s_null)
  tc(6)=a42(1,1,s_null)
  tc(7)=a42(1,2,s_null)
  tc(8)=a42(1,3,s_null)
  tc(9)=a42(2,2,s_null)
  tc(10)=a42(2,3,s_null)
  tc(11)=a42(3,3,s_null)
  tc(12)=b43(1,s_null)
  tc(13)=b43(2,s_null)
  tc(14)=b43(3,s_null)
  tc(15)=a43(1,1,1,s_null)
  tc(16)=a43(1,1,2,s_null)
  tc(17)=a43(1,1,3,s_null)
  tc(18)=a43(1,2,2,s_null)
  tc(19)=a43(1,2,3,s_null)
  tc(20)=a43(1,3,3,s_null)
  tc(21)=a43(2,2,2,s_null)
  tc(22)=a43(2,2,3,s_null)
  tc(23)=a43(2,3,3,s_null)
  tc(24)=a43(3,3,3,s_null)
  tc(25)=c44(s_null)
  tc(26)=b44(1,1,s_null)
  tc(27)=b44(1,2,s_null)
  tc(28)=b44(1,3,s_null)
  tc(29)=b44(2,2,s_null)
  tc(30)=b44(2,3,s_null)
  tc(31)=b44(3,3,s_null)
  tc(32)=a44(1,1,1,1,s_null)
  tc(33)=a44(1,1,1,2,s_null)
  tc(34)=a44(1,1,1,3,s_null)
  tc(35)=a44(1,1,2,2,s_null)
  tc(36)=a44(1,1,2,3,s_null)
  tc(37)=a44(1,1,3,3,s_null)
  tc(38)=a44(1,2,2,2,s_null)
  tc(39)=a44(1,2,2,3,s_null)
  tc(40)=a44(1,2,3,3,s_null)
  tc(41)=a44(1,3,3,3,s_null)
  tc(42)=a44(2,2,2,2,s_null)
  tc(43)=a44(2,2,2,3,s_null)
  tc(44)=a44(2,2,3,3,s_null)
  tc(45)=a44(2,3,3,3,s_null)
  tc(46)=a44(3,3,3,3,s_null)
  !
  tc(47)=a30((/ 1 /))
  tc(48)=a31(2,(/ 1 /))
  tc(49)=a31(3,(/ 1 /))
  tc(50)=b32((/ 1 /))
  tc(51)=a32(2,2,(/ 1 /))
  tc(52)=a32(2,3,(/ 1 /))
  tc(53)=a32(3,3,(/ 1 /))
  tc(54)=b33(2,(/ 1 /))
  tc(55)=b33(3,(/ 1 /))
  tc(56)=a33(2,2,2,(/ 1 /))
  tc(57)=a33(2,2,3,(/ 1 /))
  tc(58)=a33(2,3,3,(/ 1 /))
  tc(59)=a33(3,3,3,(/ 1 /))
  !
  tc(60)=a30((/ 2 /))
  tc(61)=a31(1,(/ 2 /))
  tc(62)=a31(3,(/ 2 /))
  tc(63)=b32((/ 2 /))
  tc(64)=a32(1,1,(/ 2 /))
  tc(65)=a32(1,3,(/ 2 /))
  tc(66)=a32(3,3,(/ 2 /))
  tc(67)=b33(1,(/ 2 /))
  tc(68)=b33(3,(/ 2 /))
  tc(69)=a33(1,1,1,(/ 2 /))
  tc(70)=a33(1,1,3,(/ 2 /))
  tc(71)=a33(1,3,3,(/ 2 /))
  tc(72)=a33(3,3,3,(/ 2 /))
  !
  tc(73)=a30((/ 3 /))
  tc(74)=a31(1,(/ 3 /))
  tc(75)=a31(2,(/ 3 /))
  tc(76)=b32((/ 3 /))
  tc(77)=a32(1,1,(/ 3 /))
  tc(78)=a32(1,2,(/ 3 /))
  tc(79)=a32(2,2,(/ 3 /))
  tc(80)=b33(1,(/ 3 /))
  tc(81)=b33(2,(/ 3 /))
  tc(82)=a33(1,1,1,(/ 3 /))
  tc(83)=a33(1,1,2,(/ 3 /))
  tc(84)=a33(1,2,2,(/ 3 /))
  tc(85)=a33(2,2,2,(/ 3 /))
  !
  tc(86)=a30((/ 4 /))
  tc(87)=a31(1,(/ 4 /))
  tc(88)=a31(2,(/ 4 /))
  tc(89)=b32((/ 4 /))
  tc(90)=a32(1,1,(/ 4 /))
  tc(91)=a32(1,2,(/ 4 /))
  tc(92)=a32(2,2,(/ 4 /))
  tc(93)=b33(1,(/ 4 /))
  tc(94)=b33(2,(/ 4 /))
  tc(95)=a33(1,1,1,(/ 4 /))
  tc(96)=a33(1,1,2,(/ 4 /))
  tc(97)=a33(1,2,2,(/ 4 /))
  tc(98)=a33(2,2,2,(/ 4 /))
  !
  tc(99)=a20((/ 1,3 /))
  tc(100)=a21(2,(/ 1,3 /))
  tc(101)=b22((/ 1,3 /))
  tc(102)=a22(2,2,(/ 1,3 /))
  !
  tc(103)=a20((/ 2,4 /))
  tc(104)=a21(1,(/ 2,4 /))
  tc(105)=b22((/ 2,4 /))
  tc(106)=a22(1,1,(/ 2,4 /))
  !
  g1PPPP =&
  &2.0_ki*tc(1)*S**(-1)*U**3 - 2.0_ki*tc(1)*S**(-1)*T**2*U + 2.0_ki*tc(1)&
  &*S*U - 4.0_ki*tc(2)*S**(-1)*T*U**2 - 4.0_ki*tc(2)*S**(-1)*T**2*U - 4.&
  &0_ki*tc(2)*T*U - 8.0_ki*tc(3)*S**(-2)*T*U**3 - 8.0_ki*tc(3)*S**(-2)*&
  &T**2*U**2 - 8.0_ki*tc(3)*S**(-1)*T**2*U - 8.0_ki*tc(3)*T*U - 4.0_ki*&
  &tc(4)*S**(-1)*T*U**2 - 4.0_ki*tc(4)*S**(-1)*T**2*U - 4.0_ki*tc(4)*T*&
  &U + 16.0_ki*tc(5)*S**(-2)*U**3 + 16.0_ki*tc(5)*S**(-2)*T*U**2 + 32.0_ki&
  &*tc(5)*S**(-1)*T*U - 16.0_ki*tc(5)*S**(-1)*T**2 + 16.0_ki*tc(5)*S - &
  &4.0_ki*tc(6)*T*U - 8.0_ki*tc(7)*S**(-2)*T*U**3 - 8.0_ki*tc(7)*S**(-2)*&
  &T**2*U**2 - 8.0_ki*tc(7)*S**(-1)*T**2*U - 8.0_ki*tc(7)*T*U - 8.0_ki*&
  &tc(8)*T*U - 8.0_ki*tc(9)*S**(-2)*T*U**3 - 24.0_ki*tc(9)*S**(-2)*T**2&
  &*U**2 - 8.0_ki*tc(9)*S**(-1)*T**2*U - 8.0_ki*tc(9)*T*U - 8.0_ki*tc(10)&
  &*S**(-2)*T*U**3 - 8.0_ki*tc(10)*S**(-2)*T**2*U**2 - 8.0_ki*tc(10)*&
  &S**(-1)*T**2*U - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U + 64.0_ki*tc(12&
  &)*S**(-1)*T*U + 64.0_ki*tc(13)*S**(-2)*T*U**2 - 64.0_ki*tc(13)*&
  &S**(-2)*T**2*U
  g1PPPP = g1PPPP + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)*S**(-1)&
  &*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*&
  &T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 - 64.0_ki*tc(25) - 64.0_ki&
  &*tc(25)*S**(-2)*U**2 - 64.0_ki*tc(25)*S**(-2)*T**2 + 32.0_ki*tc(26)*&
  &S**(-1)*T*U + 64.0_ki*tc(27)*S**(-2)*T*U**2 - 64.0_ki*tc(27)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(27)*S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U + &
  &64.0_ki*tc(29)*S**(-2)*T*U**2 - 64.0_ki*tc(29)*S**(-2)*T**2*U + 64.0_ki&
  &*tc(29)*S**(-1)*T*U + 64.0_ki*tc(30)*S**(-2)*T*U**2 - 64.0_ki*tc(30)&
  &*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*&
  &S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42&
  &)*S**(-2)*T**2*U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(&
  &44)*S**(-2)*T**2*U**2 - 4.0_ki*tc(47)*S**(-2)*U**3 - 8.0_ki*tc(47)*&
  &S**(-2)*T*U**2 - 4.0_ki*tc(47)*S**(-2)*T**2*U - 8.0_ki*tc(47)*&
  &S**(-1)*T*U - 4.0_ki*tc(47)*U - 16.0_ki*tc(48)*S**(-2)*T*U**2 - 8.0_ki&
  &*tc(49)*S**(-1)*T*U
  g1PPPP = g1PPPP + 32.0_ki*tc(50)*S**(-2)*U**2 - 16.0_ki*tc(51)*&
  &S**(-2)*T*U**2 - 16.0_ki*tc(52)*S**(-2)*T*U**2 - 2.0_ki*tc(60)*&
  &S**(-1)*U**2 + 2.0_ki*tc(60)*S**(-1)*T**2 - 2.0_ki*tc(60)*S + 16.0_ki*&
  &tc(63) - 2.0_ki*tc(86)*S**(-1)*U**2 + 2.0_ki*tc(86)*S**(-1)*T**2 - 2.&
  &0_ki*tc(86)*S + 16.0_ki*tc(89) - 4.0_ki*tc(99) - 4.0_ki*tc(99)*S**(-2)*&
  &U**2 + 4.0_ki*tc(99)*S**(-2)*T**2 - 4.0_ki*tc(103) - 4.0_ki*tc(103)*&
  &S**(-2)*U**2 + 4.0_ki*tc(103)*S**(-2)*T**2
  !
  g1PPPM =&
  & - 4.0_ki*tc(2)*T*U + 4.0_ki*tc(3)*S**(-1)*T*U**2 - 4.0_ki*tc(3)*&
  &S**(-1)*T**2*U - 4.0_ki*tc(3)*T*U - 4.0_ki*tc(4)*T*U + 48.0_ki*tc(5)*&
  &S**(-1)*T*U - 4.0_ki*tc(6)*T*U + 4.0_ki*tc(7)*S**(-1)*T*U**2 - 4.0_ki*&
  &tc(7)*S**(-1)*T**2*U - 8.0_ki*tc(7)*T*U - 8.0_ki*tc(8)*T*U - 16.0_ki*&
  &tc(9)*S**(-2)*T**2*U**2 + 4.0_ki*tc(9)*S**(-1)*T*U**2 - 4.0_ki*tc(9)&
  &*S**(-1)*T**2*U - 4.0_ki*tc(9)*T*U + 4.0_ki*tc(10)*S**(-1)*T*U**2 - &
  &4.0_ki*tc(10)*S**(-1)*T**2*U - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U&
  & + 64.0_ki*tc(12)*S**(-1)*T*U + 64.0_ki*tc(13)*S**(-2)*T*U**2 - 64.0_ki&
  &*tc(13)*S**(-2)*T**2*U + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)&
  &*S**(-1)*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 + 128.0_ki*tc(&
  &25)*S**(-2)*T*U + 32.0_ki*tc(26)*S**(-1)*T*U + 64.0_ki*tc(27)*&
  &S**(-2)*T*U**2 - 64.0_ki*tc(27)*S**(-2)*T**2*U + 64.0_ki*tc(27)*&
  &S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U + 64.0_ki*tc(29)*S**(-2)*T*&
  &U**2
  g1PPPM = g1PPPM - 64.0_ki*tc(29)*S**(-2)*T**2*U + 32.0_ki*tc(29)*&
  &S**(-1)*T*U + 64.0_ki*tc(30)*S**(-2)*T*U**2 - 64.0_ki*tc(30)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*S**(-1)*T*U - &
  &16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*S**(-2)*T**2*U**2&
  & - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42)*S**(-2)*T**2*&
  &U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(44)*S**(-2)*&
  &T**2*U**2 - 8.0_ki*tc(47)*S**(-1)*T*U - 8.0_ki*tc(48)*S**(-2)*T*U**2&
  & + 8.0_ki*tc(48)*S**(-2)*T**2*U - 8.0_ki*tc(48)*S**(-1)*T*U - 8.0_ki*&
  &tc(49)*S**(-1)*T*U - 32.0_ki*tc(50)*S**(-2)*T*U - 8.0_ki*tc(51)*&
  &S**(-2)*T*U**2 + 8.0_ki*tc(51)*S**(-2)*T**2*U - 8.0_ki*tc(52)*&
  &S**(-2)*T*U**2 + 8.0_ki*tc(52)*S**(-2)*T**2*U
  !
  g1PPMM =&
  & - 2.0_ki*tc(1)*S**(-1)*U**3 + 2.0_ki*tc(1)*S**(-1)*T**2*U + 2.0_ki*&
  &tc(1)*S*U + 4.0_ki*tc(2)*S**(-1)*T*U**2 + 4.0_ki*tc(2)*S**(-1)*T**2*&
  &U - 4.0_ki*tc(2)*T*U + 8.0_ki*tc(3)*S**(-2)*T*U**3 + 8.0_ki*tc(3)*&
  &S**(-2)*T**2*U**2 + 8.0_ki*tc(3)*S**(-1)*T*U**2 - 8.0_ki*tc(3)*T*U&
  & + 4.0_ki*tc(4)*S**(-1)*T*U**2 + 4.0_ki*tc(4)*S**(-1)*T**2*U - 4.0_ki*&
  &tc(4)*T*U - 16.0_ki*tc(5)*S**(-2)*U**3 - 16.0_ki*tc(5)*S**(-2)*T*&
  &U**2 - 16.0_ki*tc(5)*S**(-1)*U**2 + 32.0_ki*tc(5)*S**(-1)*T*U + 16.0_ki&
  &*tc(5)*S - 4.0_ki*tc(6)*T*U + 8.0_ki*tc(7)*S**(-2)*T*U**3 + 8.0_ki*tc(&
  &7)*S**(-2)*T**2*U**2 + 8.0_ki*tc(7)*S**(-1)*T*U**2 - 8.0_ki*tc(7)*T*&
  &U - 8.0_ki*tc(8)*T*U + 8.0_ki*tc(9)*S**(-2)*T*U**3 - 8.0_ki*tc(9)*&
  &S**(-2)*T**2*U**2 + 8.0_ki*tc(9)*S**(-1)*T*U**2 - 8.0_ki*tc(9)*T*U&
  & + 8.0_ki*tc(10)*S**(-2)*T*U**3 + 8.0_ki*tc(10)*S**(-2)*T**2*U**2 + &
  &8.0_ki*tc(10)*S**(-1)*T*U**2 - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U&
  & + 64.0_ki*tc(12)*S**(-1)*T*U + 64.0_ki*tc(13)*S**(-2)*T*U**2 - 64.0_ki&
  &*tc(13)*S**(-2)*T**2*U
  g1PPMM = g1PPMM + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)*S**(-1)&
  &*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*&
  &T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 - 64.0_ki*tc(25) - 64.0_ki&
  &*tc(25)*S**(-2)*U**2 - 64.0_ki*tc(25)*S**(-2)*T**2 + 32.0_ki*tc(26)*&
  &S**(-1)*T*U + 64.0_ki*tc(27)*S**(-2)*T*U**2 - 64.0_ki*tc(27)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(27)*S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U + &
  &64.0_ki*tc(29)*S**(-2)*T*U**2 - 64.0_ki*tc(29)*S**(-2)*T**2*U + 64.0_ki&
  &*tc(29)*S**(-1)*T*U + 64.0_ki*tc(30)*S**(-2)*T*U**2 - 64.0_ki*tc(30)&
  &*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*&
  &S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42&
  &)*S**(-2)*T**2*U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(&
  &44)*S**(-2)*T**2*U**2 + 4.0_ki*tc(47)*S**(-2)*U**3 + 8.0_ki*tc(47)*&
  &S**(-2)*T*U**2 + 4.0_ki*tc(47)*S**(-2)*T**2*U - 8.0_ki*tc(47)*&
  &S**(-1)*T*U - 4.0_ki*tc(47)*U + 16.0_ki*tc(48)*S**(-2)*T**2*U - 8.0_ki&
  &*tc(49)*S**(-1)*T*U
  g1PPMM = g1PPMM + 32.0_ki*tc(50)*S**(-2)*T**2 + 16.0_ki*tc(51)*&
  &S**(-2)*T**2*U + 16.0_ki*tc(52)*S**(-2)*T**2*U + 2.0_ki*tc(60)*&
  &S**(-1)*U**2 - 2.0_ki*tc(60)*S**(-1)*T**2 - 2.0_ki*tc(60)*S + 16.0_ki*&
  &tc(63) + 2.0_ki*tc(86)*S**(-1)*U**2 - 2.0_ki*tc(86)*S**(-1)*T**2 - 2.&
  &0_ki*tc(86)*S + 16.0_ki*tc(89) - 4.0_ki*tc(99) + 4.0_ki*tc(99)*S**(-2)*&
  &U**2 - 4.0_ki*tc(99)*S**(-2)*T**2 - 4.0_ki*tc(103) + 4.0_ki*tc(103)*&
  &S**(-2)*U**2 - 4.0_ki*tc(103)*S**(-2)*T**2
  !
  call cpu_time(t2)
  !
  ! The results are written in the file with unit 17
  !
  write (17,*) ''
  write (17,*) 'Result of graph 1:'
  write (17,*) '=================='
  write (17,*) ''
  write (17,*) 'g1PPPP ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPP%a,ki),aimag(g1PPPP%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPP%b,ki),aimag(g1PPPP%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPP%c,ki),aimag(g1PPPP%c)
  write (17,*) ''
  write (17,*) 'g1PPPM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPM%a,ki),aimag(g1PPPM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPM%b,ki),aimag(g1PPPM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPPM%c,ki),aimag(g1PPPM%c)
  write (17,*) ''
  write (17,*) 'g1PPMM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPMM%a,ki),aimag(g1PPMM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPMM%b,ki),aimag(g1PPMM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g1PPMM%c,ki),aimag(g1PPMM%c)
  write (17,*) ''  
  write (17,*) ''
  write (17,*) 'CPU time for graph 1 (all shown helicities) = ',t2-t1
  write (17,*) ''
  !
  !
  ! routine to free the allocated memory
  !
  call reset_cache()
  !
  !
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  ! GRAPH 2: 
  !
  !   k3->            <-k4 
  !   yyyyx----<----xyyyyy
  !       |         |
  !       |         |
  !       |         |
  !       |         |
  !   yyyyx---->----xyyyyy
  !   k2->            <-k1 
  !
  ! graph with reversed charge flow amounts to factor 2
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  !
  ! Definition of the S matrix
  !
  s_mat(1,1) = 0.0_ki
  s_mat(1,2) = 0.0_ki
  s_mat(1,3) = S
  s_mat(1,4) = 0.0_ki
  !
  s_mat(2,1) = 0.0_ki
  s_mat(2,2) = 0.0_ki
  s_mat(2,3) = 0.0_ki
  s_mat(2,4) = T
  !
  s_mat(3,1) = s_mat(1,3)
  s_mat(3,2) = s_mat(2,3)
  s_mat(3,3) = 0.0_ki
  s_mat(3,4) = 0.0_ki
  !
  s_mat(4,1) = s_mat(1,4)
  s_mat(4,2) = s_mat(2,4)
  s_mat(4,3) = s_mat(3,4)
  s_mat(4,4) = 0.0_ki
  !
  !
  !
  call cpu_time(t1)
  !
  ! The inverse of the S matrix is computed numerically, and also all
  ! the related quantities: b coefficients, ......
  !
  call init_invs()
  !
  tc(1)=a40(s_null)
  tc(2)=a41(1,s_null)
  tc(3)=a41(2,s_null)
  tc(4)=a41(3,s_null)
  tc(5)=b42(s_null)
  tc(6)=a42(1,1,s_null)
  tc(7)=a42(1,2,s_null)
  tc(8)=a42(1,3,s_null)
  tc(9)=a42(2,2,s_null)
  tc(10)=a42(2,3,s_null)
  tc(11)=a42(3,3,s_null)
  tc(12)=b43(1,s_null)
  tc(13)=b43(2,s_null)
  tc(14)=b43(3,s_null)
  tc(15)=a43(1,1,1,s_null)
  tc(16)=a43(1,1,2,s_null)
  tc(17)=a43(1,1,3,s_null)
  tc(18)=a43(1,2,2,s_null)
  tc(19)=a43(1,2,3,s_null)
  tc(20)=a43(1,3,3,s_null)
  tc(21)=a43(2,2,2,s_null)
  tc(22)=a43(2,2,3,s_null)
  tc(23)=a43(2,3,3,s_null)
  tc(24)=a43(3,3,3,s_null)
  tc(25)=c44(s_null)
  tc(26)=b44(1,1,s_null)
  tc(27)=b44(1,2,s_null)
  tc(28)=b44(1,3,s_null)
  tc(29)=b44(2,2,s_null)
  tc(30)=b44(2,3,s_null)
  tc(31)=b44(3,3,s_null)
  tc(32)=a44(1,1,1,1,s_null)
  tc(33)=a44(1,1,1,2,s_null)
  tc(34)=a44(1,1,1,3,s_null)
  tc(35)=a44(1,1,2,2,s_null)
  tc(36)=a44(1,1,2,3,s_null)
  tc(37)=a44(1,1,3,3,s_null)
  tc(38)=a44(1,2,2,2,s_null)
  tc(39)=a44(1,2,2,3,s_null)
  tc(40)=a44(1,2,3,3,s_null)
  tc(41)=a44(1,3,3,3,s_null)
  tc(42)=a44(2,2,2,2,s_null)
  tc(43)=a44(2,2,2,3,s_null)
  tc(44)=a44(2,2,3,3,s_null)
  tc(45)=a44(2,3,3,3,s_null)
  tc(46)=a44(3,3,3,3,s_null)
  !
  tc(47)=a30((/ 1 /))
  tc(48)=a31(2,(/ 1 /))
  tc(49)=a31(3,(/ 1 /))
  tc(50)=b32((/ 1 /))
  tc(51)=a32(2,2,(/ 1 /))
  tc(52)=a32(2,3,(/ 1 /))
  tc(53)=a32(3,3,(/ 1 /))
  tc(54)=b33(2,(/ 1 /))
  tc(55)=b33(3,(/ 1 /))
  tc(56)=a33(2,2,2,(/ 1 /))
  tc(57)=a33(2,2,3,(/ 1 /))
  tc(58)=a33(2,3,3,(/ 1 /))
  tc(59)=a33(3,3,3,(/ 1 /))
  !
  tc(60)=a30((/ 2 /))
  tc(61)=a31(1,(/ 2 /))
  tc(62)=a31(3,(/ 2 /))
  tc(63)=b32((/ 2 /))
  tc(64)=a32(1,1,(/ 2 /))
  tc(65)=a32(1,3,(/ 2 /))
  tc(66)=a32(3,3,(/ 2 /))
  tc(67)=b33(1,(/ 2 /))
  tc(68)=b33(3,(/ 2 /))
  tc(69)=a33(1,1,1,(/ 2 /))
  tc(70)=a33(1,1,3,(/ 2 /))
  tc(71)=a33(1,3,3,(/ 2 /))
  tc(72)=a33(3,3,3,(/ 2 /))
  !
  tc(73)=a30((/ 3 /))
  tc(74)=a31(1,(/ 3 /))
  tc(75)=a31(2,(/ 3 /))
  tc(76)=b32((/ 3 /))
  tc(77)=a32(1,1,(/ 3 /))
  tc(78)=a32(1,2,(/ 3 /))
  tc(79)=a32(2,2,(/ 3 /))
  tc(80)=b33(1,(/ 3 /))
  tc(81)=b33(2,(/ 3 /))
  tc(82)=a33(1,1,1,(/ 3 /))
  tc(83)=a33(1,1,2,(/ 3 /))
  tc(84)=a33(1,2,2,(/ 3 /))
  tc(85)=a33(2,2,2,(/ 3 /))
  !
  tc(86)=a30((/ 4 /))
  tc(87)=a31(1,(/ 4 /))
  tc(88)=a31(2,(/ 4 /))
  tc(89)=b32((/ 4 /))
  tc(90)=a32(1,1,(/ 4 /))
  tc(91)=a32(1,2,(/ 4 /))
  tc(92)=a32(2,2,(/ 4 /))
  tc(93)=b33(1,(/ 4 /))
  tc(94)=b33(2,(/ 4 /))
  tc(95)=a33(1,1,1,(/ 4 /))
  tc(96)=a33(1,1,2,(/ 4 /))
  tc(97)=a33(1,2,2,(/ 4 /))
  tc(98)=a33(2,2,2,(/ 4 /))
  !
  !
  tc(99)=a20((/ 1,3 /))
  tc(100)=a21(2,(/ 1,3 /))
  tc(101)=b22((/ 1,3 /))
  tc(102)=a22(2,2,(/ 1,3 /))
  !
  !
  tc(103)=a20((/ 2,4 /))
  tc(104)=a21(1,(/ 2,4 /))
  tc(105)=b22((/ 2,4 /))
  tc(106)=a22(1,1,(/ 2,4 /))
  !
  g2PPPP =&
  & - 2.0_ki*tc(1)*S**(-1)*T*U**2 + 2.0_ki*tc(1)*S**(-1)*T**3 + 2.0_ki*&
  &tc(1)*S*T - 4.0_ki*tc(2)*S**(-1)*T*U**2 - 4.0_ki*tc(2)*S**(-1)*T**2*&
  &U - 4.0_ki*tc(2)*T*U - 8.0_ki*tc(3)*S**(-2)*T**2*U**2 - 8.0_ki*tc(3)*&
  &S**(-2)*T**3*U - 8.0_ki*tc(3)*S**(-1)*T*U**2 - 8.0_ki*tc(3)*T*U - 4.0_ki&
  &*tc(4)*S**(-1)*T*U**2 - 4.0_ki*tc(4)*S**(-1)*T**2*U - 4.0_ki*tc(4)*&
  &T*U + 16.0_ki*tc(5)*S**(-2)*T**2*U + 16.0_ki*tc(5)*S**(-2)*T**3 - 16.&
  &0_ki*tc(5)*S**(-1)*U**2 + 32.0_ki*tc(5)*S**(-1)*T*U + 16.0_ki*tc(5)*S&
  & - 4.0_ki*tc(6)*T*U - 8.0_ki*tc(7)*S**(-2)*T**2*U**2 - 8.0_ki*tc(7)*&
  &S**(-2)*T**3*U - 8.0_ki*tc(7)*S**(-1)*T*U**2 - 8.0_ki*tc(7)*T*U - 8.0_ki&
  &*tc(8)*T*U - 24.0_ki*tc(9)*S**(-2)*T**2*U**2 - 8.0_ki*tc(9)*S**(-2)&
  &*T**3*U - 8.0_ki*tc(9)*S**(-1)*T*U**2 - 8.0_ki*tc(9)*T*U - 8.0_ki*tc(&
  &10)*S**(-2)*T**2*U**2 - 8.0_ki*tc(10)*S**(-2)*T**3*U - 8.0_ki*tc(10)&
  &*S**(-1)*T*U**2 - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U + 64.0_ki*tc(&
  &12)*S**(-1)*T*U - 64.0_ki*tc(13)*S**(-2)*T*U**2 + 64.0_ki*tc(13)*&
  &S**(-2)*T**2*U
  g2PPPP = g2PPPP + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)*S**(-1)&
  &*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*&
  &T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 - 64.0_ki*tc(25) - 64.0_ki&
  &*tc(25)*S**(-2)*U**2 - 64.0_ki*tc(25)*S**(-2)*T**2 + 32.0_ki*tc(26)*&
  &S**(-1)*T*U - 64.0_ki*tc(27)*S**(-2)*T*U**2 + 64.0_ki*tc(27)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(27)*S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U - &
  &64.0_ki*tc(29)*S**(-2)*T*U**2 + 64.0_ki*tc(29)*S**(-2)*T**2*U + 64.0_ki&
  &*tc(29)*S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)*T*U**2 + 64.0_ki*tc(30)&
  &*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*&
  &S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42&
  &)*S**(-2)*T**2*U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(&
  &44)*S**(-2)*T**2*U**2 - 4.0_ki*tc(47)*S**(-2)*T*U**2 - 8.0_ki*tc(47)&
  &*S**(-2)*T**2*U - 4.0_ki*tc(47)*S**(-2)*T**3 - 8.0_ki*tc(47)*S**(-1)&
  &*T*U - 4.0_ki*tc(47)*T - 16.0_ki*tc(48)*S**(-2)*T**2*U - 8.0_ki*tc(49)&
  &*S**(-1)*T*U
  g2PPPP = g2PPPP + 32.0_ki*tc(50)*S**(-2)*T**2 - 16.0_ki*tc(51)*&
  &S**(-2)*T**2*U - 16.0_ki*tc(52)*S**(-2)*T**2*U + 2.0_ki*tc(60)*&
  &S**(-1)*U**2 - 2.0_ki*tc(60)*S**(-1)*T**2 - 2.0_ki*tc(60)*S + 16.0_ki*&
  &tc(63) + 2.0_ki*tc(86)*S**(-1)*U**2 - 2.0_ki*tc(86)*S**(-1)*T**2 - 2.&
  &0_ki*tc(86)*S + 16.0_ki*tc(89) - 4.0_ki*tc(99) + 4.0_ki*tc(99)*S**(-2)*&
  &U**2 - 4.0_ki*tc(99)*S**(-2)*T**2 - 4.0_ki*tc(103) + 4.0_ki*tc(103)*&
  &S**(-2)*U**2 - 4.0_ki*tc(103)*S**(-2)*T**2
  !
  g2PPPM =&
  & - 4.0_ki*tc(2)*T*U - 4.0_ki*tc(3)*S**(-1)*T*U**2 + 4.0_ki*tc(3)*&
  &S**(-1)*T**2*U - 4.0_ki*tc(3)*T*U - 4.0_ki*tc(4)*T*U + 48.0_ki*tc(5)*&
  &S**(-1)*T*U - 4.0_ki*tc(6)*T*U - 4.0_ki*tc(7)*S**(-1)*T*U**2 + 4.0_ki*&
  &tc(7)*S**(-1)*T**2*U - 8.0_ki*tc(7)*T*U - 8.0_ki*tc(8)*T*U - 16.0_ki*&
  &tc(9)*S**(-2)*T**2*U**2 - 4.0_ki*tc(9)*S**(-1)*T*U**2 + 4.0_ki*tc(9)&
  &*S**(-1)*T**2*U - 4.0_ki*tc(9)*T*U - 4.0_ki*tc(10)*S**(-1)*T*U**2 + &
  &4.0_ki*tc(10)*S**(-1)*T**2*U - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U&
  & + 64.0_ki*tc(12)*S**(-1)*T*U - 64.0_ki*tc(13)*S**(-2)*T*U**2 + 64.0_ki&
  &*tc(13)*S**(-2)*T**2*U + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)&
  &*S**(-1)*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 + 128.0_ki*tc(&
  &25)*S**(-2)*T*U + 32.0_ki*tc(26)*S**(-1)*T*U - 64.0_ki*tc(27)*&
  &S**(-2)*T*U**2 + 64.0_ki*tc(27)*S**(-2)*T**2*U + 64.0_ki*tc(27)*&
  &S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U - 64.0_ki*tc(29)*S**(-2)*T*&
  &U**2
  g2PPPM = g2PPPM + 64.0_ki*tc(29)*S**(-2)*T**2*U + 32.0_ki*tc(29)*&
  &S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)*T*U**2 + 64.0_ki*tc(30)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*S**(-1)*T*U - &
  &16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*S**(-2)*T**2*U**2&
  & - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42)*S**(-2)*T**2*&
  &U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(44)*S**(-2)*&
  &T**2*U**2 - 8.0_ki*tc(47)*S**(-1)*T*U + 8.0_ki*tc(48)*S**(-2)*T*U**2&
  & - 8.0_ki*tc(48)*S**(-2)*T**2*U - 8.0_ki*tc(48)*S**(-1)*T*U - 8.0_ki*&
  &tc(49)*S**(-1)*T*U - 32.0_ki*tc(50)*S**(-2)*T*U + 8.0_ki*tc(51)*&
  &S**(-2)*T*U**2 - 8.0_ki*tc(51)*S**(-2)*T**2*U + 8.0_ki*tc(52)*&
  &S**(-2)*T*U**2 - 8.0_ki*tc(52)*S**(-2)*T**2*U
  !
  g2PPMM =&
  &2.0_ki*tc(1)*S**(-1)*T*U**2 - 2.0_ki*tc(1)*S**(-1)*T**3 + 2.0_ki*tc(1)&
  &*S*T + 4.0_ki*tc(2)*S**(-1)*T*U**2 + 4.0_ki*tc(2)*S**(-1)*T**2*U - 4.&
  &0_ki*tc(2)*T*U + 8.0_ki*tc(3)*S**(-2)*T**2*U**2 + 8.0_ki*tc(3)*S**(-2)&
  &*T**3*U + 8.0_ki*tc(3)*S**(-1)*T**2*U - 8.0_ki*tc(3)*T*U + 4.0_ki*tc(4&
  &)*S**(-1)*T*U**2 + 4.0_ki*tc(4)*S**(-1)*T**2*U - 4.0_ki*tc(4)*T*U - &
  &16.0_ki*tc(5)*S**(-2)*T**2*U - 16.0_ki*tc(5)*S**(-2)*T**3 + 32.0_ki*&
  &tc(5)*S**(-1)*T*U - 16.0_ki*tc(5)*S**(-1)*T**2 + 16.0_ki*tc(5)*S - 4.&
  &0_ki*tc(6)*T*U + 8.0_ki*tc(7)*S**(-2)*T**2*U**2 + 8.0_ki*tc(7)*S**(-2)&
  &*T**3*U + 8.0_ki*tc(7)*S**(-1)*T**2*U - 8.0_ki*tc(7)*T*U - 8.0_ki*tc(8&
  &)*T*U - 8.0_ki*tc(9)*S**(-2)*T**2*U**2 + 8.0_ki*tc(9)*S**(-2)*T**3*U&
  & + 8.0_ki*tc(9)*S**(-1)*T**2*U - 8.0_ki*tc(9)*T*U + 8.0_ki*tc(10)*&
  &S**(-2)*T**2*U**2 + 8.0_ki*tc(10)*S**(-2)*T**3*U + 8.0_ki*tc(10)*&
  &S**(-1)*T**2*U - 8.0_ki*tc(10)*T*U - 4.0_ki*tc(11)*T*U + 64.0_ki*tc(12&
  &)*S**(-1)*T*U - 64.0_ki*tc(13)*S**(-2)*T*U**2 + 64.0_ki*tc(13)*&
  &S**(-2)*T**2*U
  g2PPMM = g2PPMM + 64.0_ki*tc(13)*S**(-1)*T*U + 64.0_ki*tc(14)*S**(-1)&
  &*T*U - 32.0_ki*tc(18)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*&
  &T**2*U**2 - 32.0_ki*tc(22)*S**(-2)*T**2*U**2 - 64.0_ki*tc(25) - 64.0_ki&
  &*tc(25)*S**(-2)*U**2 - 64.0_ki*tc(25)*S**(-2)*T**2 + 32.0_ki*tc(26)*&
  &S**(-1)*T*U - 64.0_ki*tc(27)*S**(-2)*T*U**2 + 64.0_ki*tc(27)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(27)*S**(-1)*T*U + 64.0_ki*tc(28)*S**(-1)*T*U - &
  &64.0_ki*tc(29)*S**(-2)*T*U**2 + 64.0_ki*tc(29)*S**(-2)*T**2*U + 64.0_ki&
  &*tc(29)*S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)*T*U**2 + 64.0_ki*tc(30)&
  &*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki*tc(31)*&
  &S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(39)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42&
  &)*S**(-2)*T**2*U**2 - 32.0_ki*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(&
  &44)*S**(-2)*T**2*U**2 + 4.0_ki*tc(47)*S**(-2)*T*U**2 + 8.0_ki*tc(47)&
  &*S**(-2)*T**2*U + 4.0_ki*tc(47)*S**(-2)*T**3 - 8.0_ki*tc(47)*S**(-1)&
  &*T*U - 4.0_ki*tc(47)*T + 16.0_ki*tc(48)*S**(-2)*T*U**2 - 8.0_ki*tc(49)&
  &*S**(-1)*T*U
  g2PPMM = g2PPMM + 32.0_ki*tc(50)*S**(-2)*U**2 + 16.0_ki*tc(51)*&
  &S**(-2)*T*U**2 + 16.0_ki*tc(52)*S**(-2)*T*U**2 - 2.0_ki*tc(60)*&
  &S**(-1)*U**2 + 2.0_ki*tc(60)*S**(-1)*T**2 - 2.0_ki*tc(60)*S + 16.0_ki*&
  &tc(63) - 2.0_ki*tc(86)*S**(-1)*U**2 + 2.0_ki*tc(86)*S**(-1)*T**2 - 2.&
  &0_ki*tc(86)*S + 16.0_ki*tc(89) - 4.0_ki*tc(99) - 4.0_ki*tc(99)*S**(-2)*&
  &U**2 + 4.0_ki*tc(99)*S**(-2)*T**2 - 4.0_ki*tc(103) - 4.0_ki*tc(103)*&
  &S**(-2)*U**2 + 4.0_ki*tc(103)*S**(-2)*T**2
  !
  call cpu_time(t2)
  !
  ! The results are written in the file with unit 17
  !
  write (17,*) ''
  write (17,*) 'Result of graph 2:'  
  write (17,*) '=================='
  write (17,*) ''
  write (17,*) 'g2PPPP ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPP%a,ki),aimag(g2PPPP%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPP%b,ki),aimag(g2PPPP%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPP%c,ki),aimag(g2PPPP%c)
  write (17,*) ''
  write (17,*) 'g2PPPM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPM%a,ki),aimag(g2PPPM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPM%b,ki),aimag(g2PPPM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPPM%c,ki),aimag(g2PPPM%c)
  write (17,*) ''
  write (17,*) 'g2PPMM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPMM%a,ki),aimag(g2PPMM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPMM%b,ki),aimag(g2PPMM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g2PPMM%c,ki),aimag(g2PPMM%c)
  write (17,*) ''  
  write (17,*) ''
  write (17,*) 'CPU time for graph 2 = ',t2-t1
  write (17,*) ''
  !
  ! routine to free the allocated memory
  !
  call reset_cache()
  !
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  ! GRAPH 3: 
  !
  !   k2->            <-k4 
  !   yyyyx----<----xyyyyy
  !       |         |
  !       |         |
  !       |         |
  !       |         |
  !   yyyyx---->----xyyyyy
  !   k3->            <-k1 
  ! 
  ! graph with reversed charge flow amounts to factor 2
  !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
  !
  ! Definition of the S matrix
  !
  s_mat(1,1) = 0.0_ki
  s_mat(1,2) = 0.0_ki
  s_mat(1,3) = U
  s_mat(1,4) = 0.0_ki
  !
  s_mat(2,1) = 0.0_ki
  s_mat(2,2) = 0.0_ki
  s_mat(2,3) = 0.0_ki
  s_mat(2,4) = T
  !
  s_mat(3,1) = s_mat(1,3)
  s_mat(3,2) = s_mat(2,3)
  s_mat(3,3) = 0.0_ki
  s_mat(3,4) = 0.0_ki
  !
  s_mat(4,1) = s_mat(1,4)
  s_mat(4,2) = s_mat(2,4)
  s_mat(4,3) = s_mat(3,4)
  s_mat(4,4) = 0.0_ki
  !
  call cpu_time(t1)
  !
  ! The inverse of the S matrix is computed numerically, and also all
  ! the related quantities: b coefficients, ......
  !
  call init_invs()
  !
  tc(1)=a40(s_null)
  tc(2)=a41(1,s_null)
  tc(3)=a41(2,s_null)
  tc(4)=a41(3,s_null)
  tc(5)=b42(s_null)
  tc(6)=a42(1,1,s_null)
  tc(7)=a42(1,2,s_null)
  tc(8)=a42(1,3,s_null)
  tc(9)=a42(2,2,s_null)
  tc(10)=a42(2,3,s_null)
  tc(11)=a42(3,3,s_null)
  tc(12)=b43(1,s_null)
  tc(13)=b43(2,s_null)
  tc(14)=b43(3,s_null)
  tc(15)=a43(1,1,1,s_null)
  tc(16)=a43(1,1,2,s_null)
  tc(17)=a43(1,1,3,s_null)
  tc(18)=a43(1,2,2,s_null)
  tc(19)=a43(1,2,3,s_null)
  tc(20)=a43(1,3,3,s_null)
  tc(21)=a43(2,2,2,s_null)
  tc(22)=a43(2,2,3,s_null)
  tc(23)=a43(2,3,3,s_null)
  tc(24)=a43(3,3,3,s_null)
  tc(25)=c44(s_null)
  tc(26)=b44(1,1,s_null)
  tc(27)=b44(1,2,s_null)
  tc(28)=b44(1,3,s_null)
  tc(29)=b44(2,2,s_null)
  tc(30)=b44(2,3,s_null)
  tc(31)=b44(3,3,s_null)
  tc(32)=a44(1,1,1,1,s_null)
  tc(33)=a44(1,1,1,2,s_null)
  tc(34)=a44(1,1,1,3,s_null)
  tc(35)=a44(1,1,2,2,s_null)
  tc(36)=a44(1,1,2,3,s_null)
  tc(37)=a44(1,1,3,3,s_null)
  tc(38)=a44(1,2,2,2,s_null)
  tc(39)=a44(1,2,2,3,s_null)
  tc(40)=a44(1,2,3,3,s_null)
  tc(41)=a44(1,3,3,3,s_null)
  tc(42)=a44(2,2,2,2,s_null)
  tc(43)=a44(2,2,2,3,s_null)
  tc(44)=a44(2,2,3,3,s_null)
  tc(45)=a44(2,3,3,3,s_null)
  tc(46)=a44(3,3,3,3,s_null)
  !
  tc(47)=a30((/ 1 /))
  tc(48)=a31(2,(/ 1 /))
  tc(49)=a31(3,(/ 1 /))
  tc(50)=b32((/ 1 /))
  tc(51)=a32(2,2,(/ 1 /))
  tc(52)=a32(2,3,(/ 1 /))
  tc(53)=a32(3,3,(/ 1 /))
  tc(54)=b33(2,(/ 1 /))
  tc(55)=b33(3,(/ 1 /))
  tc(56)=a33(2,2,2,(/ 1 /))
  tc(57)=a33(2,2,3,(/ 1 /))
  tc(58)=a33(2,3,3,(/ 1 /))
  tc(59)=a33(3,3,3,(/ 1 /))
  !
  tc(60)=a30((/ 2 /))
  tc(61)=a31(1,(/ 2 /))
  tc(62)=a31(3,(/ 2 /))
  tc(63)=b32((/ 2 /))
  tc(64)=a32(1,1,(/ 2 /))
  tc(65)=a32(1,3,(/ 2 /))
  tc(66)=a32(3,3,(/ 2 /))
  tc(67)=b33(1,(/ 2 /))
  tc(68)=b33(3,(/ 2 /))
  tc(69)=a33(1,1,1,(/ 2 /))
  tc(70)=a33(1,1,3,(/ 2 /))
  tc(71)=a33(1,3,3,(/ 2 /))
  tc(72)=a33(3,3,3,(/ 2 /))
  !
  tc(73)=a30((/ 3 /))
  tc(74)=a31(1,(/ 3 /))
  tc(75)=a31(2,(/ 3 /))
  tc(76)=b32((/ 3 /))
  tc(77)=a32(1,1,(/ 3 /))
  tc(78)=a32(1,2,(/ 3 /))
  tc(79)=a32(2,2,(/ 3 /))
  tc(80)=b33(1,(/ 3 /))
  tc(81)=b33(2,(/ 3 /))
  tc(82)=a33(1,1,1,(/ 3 /))
  tc(83)=a33(1,1,2,(/ 3 /))
  tc(84)=a33(1,2,2,(/ 3 /))
  tc(85)=a33(2,2,2,(/ 3 /))
  !
  tc(86)=a30((/ 4 /))
  tc(87)=a31(1,(/ 4 /))
  tc(88)=a31(2,(/ 4 /))
  tc(89)=b32((/ 4 /))
  tc(90)=a32(1,1,(/ 4 /))
  tc(91)=a32(1,2,(/ 4 /))
  tc(92)=a32(2,2,(/ 4 /))
  tc(93)=b33(1,(/ 4 /))
  tc(94)=b33(2,(/ 4 /))
  tc(95)=a33(1,1,1,(/ 4 /))
  tc(96)=a33(1,1,2,(/ 4 /))
  tc(97)=a33(1,2,2,(/ 4 /))
  tc(98)=a33(2,2,2,(/ 4 /))
  !
  tc(99)=a20((/ 1,3 /))
  tc(100)=a21(2,(/ 1,3 /))
  tc(101)=b22((/ 1,3 /))
  tc(102)=a22(2,2,(/ 1,3 /))
  !
  tc(103)=a20((/ 2,4 /))
  tc(104)=a21(1,(/ 2,4 /))
  tc(105)=b22((/ 2,4 /))
  tc(106)=a22(1,1,(/ 2,4 /))
  !
  !
  g3PPPP =&
  &2.0_ki*tc(1)*S**(-2)*T*U**3 - 2.0_ki*tc(1)*S**(-2)*T**3*U - 2.0_ki*tc(&
  &1)*T*U + 4.0_ki*tc(2)*S**(-2)*T*U**3 - 4.0_ki*tc(2)*S**(-2)*T**3*U&
  & + 4.0_ki*tc(2)*S**(-1)*T*U**2 + 4.0_ki*tc(2)*S**(-1)*T**2*U + 8.0_ki*&
  &tc(3)*S**(-2)*T*U**3 - 8.0_ki*tc(3)*S**(-2)*T**3*U + 8.0_ki*tc(3)*&
  &S**(-1)*T*U**2 + 8.0_ki*tc(3)*S**(-1)*T**2*U + 4.0_ki*tc(4)*S**(-2)*&
  &T*U**3 - 4.0_ki*tc(4)*S**(-2)*T**3*U + 4.0_ki*tc(4)*S**(-1)*T*U**2&
  & + 4.0_ki*tc(4)*S**(-1)*T**2*U + 16.0_ki*tc(5)*S**(-2)*U**3 - 16.0_ki*&
  &tc(5)*S**(-2)*T*U**2 + 16.0_ki*tc(5)*S**(-2)*T**2*U + 16.0_ki*tc(5)*&
  &S**(-2)*T**3 - 16.0_ki*tc(5)*U - 16.0_ki*tc(5)*T + 4.0_ki*tc(6)*&
  &S**(-1)*T*U**2 + 4.0_ki*tc(6)*S**(-1)*T**2*U + 8.0_ki*tc(7)*S**(-2)*&
  &T*U**3 - 16.0_ki*tc(7)*S**(-2)*T**2*U**2 - 8.0_ki*tc(7)*S**(-2)*T**3&
  &*U + 8.0_ki*tc(7)*S**(-1)*T*U**2 + 8.0_ki*tc(7)*S**(-1)*T**2*U + 8.0_ki&
  &*tc(8)*S**(-2)*T*U**3 - 16.0_ki*tc(8)*S**(-2)*T**2*U**2 - 8.0_ki*tc(&
  &8)*S**(-2)*T**3*U + 8.0_ki*tc(9)*S**(-2)*T*U**3 - 16.0_ki*tc(9)*&
  &S**(-2)*T**2*U**2
  g3PPPP = g3PPPP - 8.0_ki*tc(9)*S**(-2)*T**3*U + 8.0_ki*tc(9)*S**(-1)*&
  &T*U**2 + 8.0_ki*tc(9)*S**(-1)*T**2*U + 8.0_ki*tc(10)*S**(-2)*T*U**3&
  & - 16.0_ki*tc(10)*S**(-2)*T**2*U**2 - 8.0_ki*tc(10)*S**(-2)*T**3*U&
  & + 8.0_ki*tc(10)*S**(-1)*T*U**2 + 8.0_ki*tc(10)*S**(-1)*T**2*U + 4.0_ki&
  &*tc(11)*S**(-1)*T*U**2 + 4.0_ki*tc(11)*S**(-1)*T**2*U - 32.0_ki*tc(&
  &12)*S**(-2)*T*U**2 + 32.0_ki*tc(12)*S**(-2)*T**2*U + 32.0_ki*tc(12)*&
  &S**(-1)*T*U - 64.0_ki*tc(13)*S**(-2)*T*U**2 + 64.0_ki*tc(13)*S**(-2)&
  &*T**2*U + 64.0_ki*tc(13)*S**(-1)*T*U - 32.0_ki*tc(14)*S**(-2)*T*U**2&
  & + 32.0_ki*tc(14)*S**(-2)*T**2*U + 32.0_ki*tc(14)*S**(-1)*T*U - 16.0_ki&
  &*tc(16)*S**(-2)*T**2*U**2 - 16.0_ki*tc(17)*S**(-2)*T**2*U**2 - 48.0_ki&
  &*tc(18)*S**(-2)*T**2*U**2 - 64.0_ki*tc(19)*S**(-2)*T**2*U**2 - 16.&
  &0_ki*tc(20)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*T**2*U**2 - &
  &48.0_ki*tc(22)*S**(-2)*T**2*U**2 - 16.0_ki*tc(23)*S**(-2)*T**2*U**2&
  & - 64.0_ki*tc(25) - 64.0_ki*tc(25)*S**(-2)*U**2 - 64.0_ki*tc(25)*&
  &S**(-2)*T**2 + 32.0_ki*tc(26)*S**(-1)*T*U - 64.0_ki*tc(27)*S**(-2)*T&
  &*U**2
  g3PPPP = g3PPPP + 64.0_ki*tc(27)*S**(-2)*T**2*U + 64.0_ki*tc(27)*&
  &S**(-1)*T*U - 64.0_ki*tc(28)*S**(-2)*T*U**2 + 64.0_ki*tc(28)*S**(-2)&
  &*T**2*U - 64.0_ki*tc(29)*S**(-2)*T*U**2 + 64.0_ki*tc(29)*S**(-2)*&
  &T**2*U + 64.0_ki*tc(29)*S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)*T*U**2&
  & + 64.0_ki*tc(30)*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U + 32.0_ki&
  &*tc(31)*S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - 32.0_ki*tc(&
  &36)*S**(-2)*T**2*U**2 - 16.0_ki*tc(37)*S**(-2)*T**2*U**2 - 32.0_ki*&
  &tc(38)*S**(-2)*T**2*U**2 - 64.0_ki*tc(39)*S**(-2)*T**2*U**2 - 32.0_ki&
  &*tc(40)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42)*S**(-2)*T**2*U**2 - 32.0_ki&
  &*tc(43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(44)*S**(-2)*T**2*U**2 + 4.0_ki&
  &*tc(47)*S**(-2)*T*U**2 - 4.0_ki*tc(47)*S**(-2)*T**3 + 8.0_ki*tc(47)&
  &*S**(-1)*T*U + 4.0_ki*tc(47)*T - 16.0_ki*tc(48)*S**(-2)*T**2*U + 8.0_ki&
  &*tc(49)*S**(-1)*T*U + 32.0_ki*tc(50)*S**(-2)*T**2 - 16.0_ki*tc(51)*&
  &S**(-2)*T**2*U - 16.0_ki*tc(52)*S**(-2)*T**2*U - 2.0_ki*tc(60)*&
  &S**(-2)*U**3 + 2.0_ki*tc(60)*S**(-2)*T**2*U + 2.0_ki*tc(60)*U - 4.0_ki&
  &*tc(61)*S**(-1)*T*U
  g3PPPP = g3PPPP - 4.0_ki*tc(62)*S**(-1)*T*U + 16.0_ki*tc(63)*S**(-2)*&
  &U**2 + 8.0_ki*tc(65)*S**(-2)*T*U**2 - 2.0_ki*tc(86)*S**(-2)*U**3 + 2.&
  &0_ki*tc(86)*S**(-2)*T**2*U + 4.0_ki*tc(86)*S**(-1)*T*U + 2.0_ki*tc(86)&
  &*U - 8.0_ki*tc(87)*S**(-2)*T*U**2 + 4.0_ki*tc(88)*S**(-1)*T*U + 16.0_ki&
  &*tc(89)*S**(-2)*U**2 - 8.0_ki*tc(90)*S**(-2)*T*U**2 - 8.0_ki*tc(91)*&
  &S**(-2)*T*U**2 + 4.0_ki*tc(99) - 4.0_ki*tc(99)*S**(-2)*U**2 - 4.0_ki*&
  &tc(99)*S**(-2)*T**2 + 4.0_ki*tc(103) - 4.0_ki*tc(103)*S**(-2)*U**2&
  & - 4.0_ki*tc(103)*S**(-2)*T**2
  !
  g3PPPM =&
  & - 4.0_ki*tc(1)*S**(-2)*T**2*U**2 - 4.0_ki*tc(1)*S**(-2)*T**3*U + 4.0_ki&
  &*tc(2)*S**(-2)*T*U**3 - 4.0_ki*tc(2)*S**(-2)*T**3*U + 4.0_ki*tc(3)*&
  &S**(-2)*T*U**3 - 4.0_ki*tc(3)*S**(-2)*T**2*U**2 - 8.0_ki*tc(3)*&
  &S**(-2)*T**3*U + 4.0_ki*tc(3)*S**(-1)*T*U**2 + 4.0_ki*tc(3)*S**(-1)*&
  &T**2*U - 4.0_ki*tc(4)*S**(-2)*T**2*U**2 - 4.0_ki*tc(4)*S**(-2)*T**3*&
  &U + 4.0_ki*tc(4)*S**(-1)*T*U**2 + 4.0_ki*tc(4)*S**(-1)*T**2*U - 48.0_ki&
  &*tc(5)*S**(-2)*T*U**2 - 16.0_ki*tc(5)*S**(-2)*T**2*U + 4.0_ki*tc(7)*&
  &S**(-2)*T*U**3 - 16.0_ki*tc(7)*S**(-2)*T**2*U**2 - 4.0_ki*tc(7)*&
  &S**(-2)*T**3*U + 4.0_ki*tc(8)*S**(-2)*T*U**3 - 16.0_ki*tc(8)*S**(-2)&
  &*T**2*U**2 - 4.0_ki*tc(8)*S**(-2)*T**3*U + 4.0_ki*tc(9)*S**(-2)*T*&
  &U**3 - 16.0_ki*tc(9)*S**(-2)*T**2*U**2 - 4.0_ki*tc(9)*S**(-2)*T**3*U&
  & + 4.0_ki*tc(9)*S**(-1)*T*U**2 + 4.0_ki*tc(9)*S**(-1)*T**2*U + 4.0_ki*&
  &tc(10)*S**(-2)*T*U**3 - 16.0_ki*tc(10)*S**(-2)*T**2*U**2 - 4.0_ki*&
  &tc(10)*S**(-2)*T**3*U + 8.0_ki*tc(10)*S**(-1)*T*U**2 + 8.0_ki*tc(10)&
  &*S**(-1)*T**2*U
  g3PPPM = g3PPPM + 4.0_ki*tc(11)*S**(-1)*T*U**2 + 4.0_ki*tc(11)*&
  &S**(-1)*T**2*U - 32.0_ki*tc(12)*S**(-2)*T*U**2 + 32.0_ki*tc(12)*&
  &S**(-2)*T**2*U - 96.0_ki*tc(13)*S**(-2)*T*U**2 + 32.0_ki*tc(13)*&
  &S**(-2)*T**2*U + 32.0_ki*tc(13)*S**(-1)*T*U - 64.0_ki*tc(14)*S**(-2)&
  &*T*U**2 + 32.0_ki*tc(14)*S**(-1)*T*U - 16.0_ki*tc(16)*S**(-2)*T**2*&
  &U**2 - 16.0_ki*tc(17)*S**(-2)*T**2*U**2 - 48.0_ki*tc(18)*S**(-2)*&
  &T**2*U**2 - 64.0_ki*tc(19)*S**(-2)*T**2*U**2 - 16.0_ki*tc(20)*&
  &S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*S**(-2)*T**2*U**2 - 48.0_ki*tc(22&
  &)*S**(-2)*T**2*U**2 - 16.0_ki*tc(23)*S**(-2)*T**2*U**2 + 128.0_ki*&
  &tc(25)*S**(-2)*T*U - 64.0_ki*tc(27)*S**(-2)*T*U**2 + 64.0_ki*tc(27)*&
  &S**(-2)*T**2*U - 64.0_ki*tc(28)*S**(-2)*T*U**2 + 64.0_ki*tc(28)*&
  &S**(-2)*T**2*U - 64.0_ki*tc(29)*S**(-2)*T*U**2 + 64.0_ki*tc(29)*&
  &S**(-2)*T**2*U + 32.0_ki*tc(29)*S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)&
  &*T*U**2 + 64.0_ki*tc(30)*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U&
  & + 32.0_ki*tc(31)*S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - &
  &32.0_ki*tc(36)*S**(-2)*T**2*U**2
  g3PPPM = g3PPPM - 16.0_ki*tc(37)*S**(-2)*T**2*U**2 - 32.0_ki*tc(38)*&
  &S**(-2)*T**2*U**2 - 64.0_ki*tc(39)*S**(-2)*T**2*U**2 - 32.0_ki*tc(40&
  &)*S**(-2)*T**2*U**2 - 16.0_ki*tc(42)*S**(-2)*T**2*U**2 - 32.0_ki*tc(&
  &43)*S**(-2)*T**2*U**2 - 16.0_ki*tc(44)*S**(-2)*T**2*U**2 + 4.0_ki*&
  &tc(47)*S**(-2)*T*U**2 + 4.0_ki*tc(47)*S**(-2)*T**2*U + 4.0_ki*tc(47)&
  &*S**(-1)*T*U + 8.0_ki*tc(48)*S**(-2)*T*U**2 - 8.0_ki*tc(48)*S**(-2)*&
  &T**2*U + 4.0_ki*tc(49)*S**(-1)*T*U - 32.0_ki*tc(50)*S**(-2)*T*U + 8.0_ki&
  &*tc(51)*S**(-2)*T*U**2 - 8.0_ki*tc(51)*S**(-2)*T**2*U + 8.0_ki*tc(&
  &52)*S**(-2)*T*U**2 - 8.0_ki*tc(52)*S**(-2)*T**2*U + 4.0_ki*tc(60)*&
  &S**(-2)*T*U**2 + 4.0_ki*tc(60)*S**(-2)*T**2*U - 4.0_ki*tc(62)*&
  &S**(-1)*T*U - 16.0_ki*tc(63)*S**(-2)*T*U - 8.0_ki*tc(65)*S**(-2)*&
  &T**2*U + 4.0_ki*tc(86)*S**(-1)*T*U - 8.0_ki*tc(87)*S**(-2)*T*U**2 + &
  &4.0_ki*tc(87)*S**(-1)*T*U + 4.0_ki*tc(88)*S**(-1)*T*U - 16.0_ki*tc(89)&
  &*S**(-2)*T*U - 8.0_ki*tc(90)*S**(-2)*T*U**2 - 8.0_ki*tc(91)*S**(-2)*&
  &T*U**2 + 8.0_ki*tc(99)*S**(-2)*T*U + 8.0_ki*tc(103)*S**(-2)*T*U
  !
  g3PPMM =&
  &2.0_ki*tc(1)*S**(-2)*T*U**3 + 4.0_ki*tc(1)*S**(-2)*T**2*U**2 + 2.0_ki*&
  &tc(1)*S**(-2)*T**3*U - 2.0_ki*tc(1)*T*U + 4.0_ki*tc(2)*S**(-1)*T*&
  &U**2 + 4.0_ki*tc(2)*S**(-1)*T**2*U + 8.0_ki*tc(3)*S**(-1)*T*U**2 + 8.&
  &0_ki*tc(3)*S**(-1)*T**2*U + 4.0_ki*tc(4)*S**(-1)*T*U**2 + 4.0_ki*tc(4)&
  &*S**(-1)*T**2*U + 32.0_ki*tc(5)*S**(-2)*T**2*U - 16.0_ki*tc(5)*U - &
  &16.0_ki*tc(5)*T + 4.0_ki*tc(6)*S**(-1)*T*U**2 + 4.0_ki*tc(6)*S**(-1)*&
  &T**2*U - 16.0_ki*tc(7)*S**(-2)*T**2*U**2 + 8.0_ki*tc(7)*S**(-1)*T*&
  &U**2 + 8.0_ki*tc(7)*S**(-1)*T**2*U - 16.0_ki*tc(8)*S**(-2)*T**2*U**2&
  & - 16.0_ki*tc(9)*S**(-2)*T**2*U**2 + 8.0_ki*tc(9)*S**(-1)*T*U**2 + 8.&
  &0_ki*tc(9)*S**(-1)*T**2*U - 16.0_ki*tc(10)*S**(-2)*T**2*U**2 + 8.0_ki*&
  &tc(10)*S**(-1)*T*U**2 + 8.0_ki*tc(10)*S**(-1)*T**2*U + 4.0_ki*tc(11)&
  &*S**(-1)*T*U**2 + 4.0_ki*tc(11)*S**(-1)*T**2*U - 32.0_ki*tc(12)*&
  &S**(-2)*T*U**2 + 32.0_ki*tc(12)*S**(-2)*T**2*U + 32.0_ki*tc(12)*&
  &S**(-1)*T*U - 64.0_ki*tc(13)*S**(-2)*T*U**2 + 64.0_ki*tc(13)*S**(-2)&
  &*T**2*U
  g3PPMM = g3PPMM + 64.0_ki*tc(13)*S**(-1)*T*U - 32.0_ki*tc(14)*S**(-2)&
  &*T*U**2 + 32.0_ki*tc(14)*S**(-2)*T**2*U + 32.0_ki*tc(14)*S**(-1)*T*U&
  & - 16.0_ki*tc(16)*S**(-2)*T**2*U**2 - 16.0_ki*tc(17)*S**(-2)*T**2*&
  &U**2 - 48.0_ki*tc(18)*S**(-2)*T**2*U**2 - 64.0_ki*tc(19)*S**(-2)*&
  &T**2*U**2 - 16.0_ki*tc(20)*S**(-2)*T**2*U**2 - 32.0_ki*tc(21)*&
  &S**(-2)*T**2*U**2 - 48.0_ki*tc(22)*S**(-2)*T**2*U**2 - 16.0_ki*tc(23&
  &)*S**(-2)*T**2*U**2 - 64.0_ki*tc(25) - 64.0_ki*tc(25)*S**(-2)*U**2&
  & - 64.0_ki*tc(25)*S**(-2)*T**2 + 32.0_ki*tc(26)*S**(-1)*T*U - 64.0_ki*&
  &tc(27)*S**(-2)*T*U**2 + 64.0_ki*tc(27)*S**(-2)*T**2*U + 64.0_ki*tc(&
  &27)*S**(-1)*T*U - 64.0_ki*tc(28)*S**(-2)*T*U**2 + 64.0_ki*tc(28)*&
  &S**(-2)*T**2*U - 64.0_ki*tc(29)*S**(-2)*T*U**2 + 64.0_ki*tc(29)*&
  &S**(-2)*T**2*U + 64.0_ki*tc(29)*S**(-1)*T*U - 64.0_ki*tc(30)*S**(-2)&
  &*T*U**2 + 64.0_ki*tc(30)*S**(-2)*T**2*U + 64.0_ki*tc(30)*S**(-1)*T*U&
  & + 32.0_ki*tc(31)*S**(-1)*T*U - 16.0_ki*tc(35)*S**(-2)*T**2*U**2 - &
  &32.0_ki*tc(36)*S**(-2)*T**2*U**2 - 16.0_ki*tc(37)*S**(-2)*T**2*U**2&
  & - 32.0_ki*tc(38)*S**(-2)*T**2*U**2
  g3PPMM = g3PPMM - 64.0_ki*tc(39)*S**(-2)*T**2*U**2 - 32.0_ki*tc(40)*&
  &S**(-2)*T**2*U**2 - 16.0_ki*tc(42)*S**(-2)*T**2*U**2 - 32.0_ki*tc(43&
  &)*S**(-2)*T**2*U**2 - 16.0_ki*tc(44)*S**(-2)*T**2*U**2 - 4.0_ki*tc(&
  &47)*S**(-2)*T*U**2 - 8.0_ki*tc(47)*S**(-2)*T**2*U - 4.0_ki*tc(47)*&
  &S**(-2)*T**3 + 8.0_ki*tc(47)*S**(-1)*T*U + 4.0_ki*tc(47)*T + 16.0_ki*&
  &tc(48)*S**(-2)*T*U**2 + 8.0_ki*tc(49)*S**(-1)*T*U + 32.0_ki*tc(50)*&
  &S**(-2)*U**2 + 16.0_ki*tc(51)*S**(-2)*T*U**2 + 16.0_ki*tc(52)*&
  &S**(-2)*T*U**2 - 2.0_ki*tc(60)*S**(-2)*U**3 - 4.0_ki*tc(60)*S**(-2)*&
  &T*U**2 - 2.0_ki*tc(60)*S**(-2)*T**2*U + 2.0_ki*tc(60)*U - 4.0_ki*tc(61&
  &)*S**(-1)*T*U - 4.0_ki*tc(62)*S**(-1)*T*U + 16.0_ki*tc(63)*S**(-2)*&
  &T**2 - 8.0_ki*tc(65)*S**(-2)*T**2*U - 2.0_ki*tc(86)*S**(-2)*U**3 - 4.&
  &0_ki*tc(86)*S**(-2)*T*U**2 - 2.0_ki*tc(86)*S**(-2)*T**2*U + 4.0_ki*tc(&
  &86)*S**(-1)*T*U + 2.0_ki*tc(86)*U + 8.0_ki*tc(87)*S**(-2)*T**2*U + 4.&
  &0_ki*tc(88)*S**(-1)*T*U + 16.0_ki*tc(89)*S**(-2)*T**2 + 8.0_ki*tc(90)*&
  &S**(-2)*T**2*U + 8.0_ki*tc(91)*S**(-2)*T**2*U + 4.0_ki*tc(99) - 4.0_ki&
  &*tc(99)*S**(-2)*U**2
  g3PPMM = g3PPMM - 4.0_ki*tc(99)*S**(-2)*T**2 + 4.0_ki*tc(103) - 4.0_ki*&
  &tc(103)*S**(-2)*U**2 - 4.0_ki*tc(103)*S**(-2)*T**2
  !
  ! routine to free the allocated memory
  !
  call clear_cache()
  !
  call deallocation_s()
  !
  call cpu_time(t2)
  !
  ! The results are written in the file with unit 17
  !
  write (17,*) ''
  write (17,*) 'Result of graph 3:'
  write (17,*) '=================='
  write (17,*) ''
  write (17,*) 'g3PPPP ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPP%a,ki),aimag(g3PPPP%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPP%b,ki),aimag(g3PPPP%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPP%c,ki),aimag(g3PPPP%c)
  write (17,*) ''
  write (17,*) 'g3PPPM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPM%a,ki),aimag(g3PPPM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPM%b,ki),aimag(g3PPPM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPPM%c,ki),aimag(g3PPPM%c)
  write (17,*) ''
  write (17,*) 'g3PPMM ='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPMM%a,ki),aimag(g3PPMM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPMM%b,ki),aimag(g3PPMM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(g3PPMM%c,ki),aimag(g3PPMM%c)
  write (17,*) ''  
  write (17,*) ''
  write (17,*) 'CPU time for graph 3 = ',t2-t1
  write (17,*) ''
  !
  ampPPPP = g1PPPP+g2PPPP+g3PPPP
  ampPPPM = g1PPPM+g2PPPM+g3PPPM
  ampPPMM = g1PPMM+g2PPMM+g3PPMM
  !
  ! analytical result for ++--
  !
  resPPMM = -8.0_ki*(1.0_ki+(T-U)/S*log(T/U)+(T**2+U**2)/2/S**2*(pi**2+log(T/U)**2))
  !
  ! The results are written in the file with unit 17
  !
  write (17,*) ''
  write (17,*) 'Four photon helicity amplitudes'
  write (17,*) '==============================='
  write (17,*) ''
  write (17,*) 'amp(++++)='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPP%a,ki),aimag(ampPPPP%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPP%b,ki),aimag(ampPPPP%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPP%c,ki),aimag(ampPPPP%c)
  write (17,*) ''
  write (17,*) 'amp(+++-)='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPM%a,ki),aimag(ampPPPM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPM%b,ki),aimag(ampPPPM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPPM%c,ki),aimag(ampPPPM%c)
  write (17,*) ''
  write (17,*) 'amp(++--)='
  write (17,'("+ 1/epsilon^2 * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPMM%a,ki),aimag(ampPPMM%a)
  write (17,'("+ 1/epsilon   * (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPMM%b,ki),aimag(ampPPMM%b)
  write (17,'("+               (",e16.10,1x,"+ I*",1x,e16.10,")")') real(ampPPMM%c,ki),aimag(ampPPMM%c)
  write (17,*) ''
  write (17,*) 'Analytical result:'   
  write (17,*) '=================='
  write (17,*) ''  
  write (17,*) 'amp(++++) = +8'  
  write (17,*) 'amp(+++-) = -8'
  !
  if (rat_or_tot_par == 'rat') then
    !
    write (17,*) 'amp(++--) = -8'
    !
  else
    !
    write (17,*) 'amp(++--) = -8*( + 1                                           '
    write (17,*) '                 + (t-u)/s*log(t/u)                            '
    write (17,*) '                 + (t**2+u**2)/2/s**2*( pi**2 + log(t/u)**2 ) )'
    write (17,'("           =",e16.10)') resPPMM
    !
  end if
  !
  write (17,*) ''  
  write (6,*) 'The result has been written to result-4photon.txt'
  close(17)
  close(19)
  !
end program main
!
!
