! 
!****h* src/module/equal
! NAME
!
!  Module equal
!
! USAGE
!
!  use equal
!
! DESCRIPTION
!
!  This module is used to compare two objects
!
! OUTPUT
!
!  This module exports one function:
!  * equal_real -- to compare two reals
!
! USES
!
!  * precision (src/module/preci_double.f90)
!
!
!*****
module equal
  !
  use precision
  implicit none
  !
  contains
    !
    !****f* src/module/equal/equal_real
    ! NAME
    !
    !  Function equal_real
    !
    ! USAGE
    !
    !  logical = equal_real(xa,xb)
    !
    ! DESCRIPTION
    !
    !  This function compares two real (of same kind) by computing their 
    !  difference and compares it to epsilon (of the same kind)
    !  true if | xa - xb | <= epsilon, false otherwise
    !
    ! INPUTS
    !
    !  * xa -- a real of type ki
    !  * xb -- a real of type ki
    !
    ! SIDE EFFECTS
    !
    !  No side effect
    !
    ! RETURN VALUE
    !
    !  It returns a logical
    !
    ! EXAMPLE
    !
    !
    !*****
    function equal_real(xa,xb)
      !
      real(ki), intent(in) :: xa,xb
      logical equal_real
      !
      real(ki) :: my_epsilon
      !
      my_epsilon = epsilon(1._ki)
      equal_real = abs(xa-xb) <= my_epsilon
      !
    end function equal_real
    !
end module equal
