! 
!****h* src/module/multiply_div
! NAME
!
!  Module multiply_div
!
! USAGE
!
!  use multiply_div
!
! DESCRIPTION
!
!  This module contains the function mult_div, This function computes 
!  numericaly (1+alpha*epsilon)*(A/epsilon+B)
!
! OUTPUT
!
!  This module exports the function mult_div
!
!
! USES
!
!  * precision (src/module/preci_double.f90)
!
!*****
module multiply_div
  !
  use precision
  implicit none
  !
  contains
    !
    !****f* src/module/multiply_div/mult_div
    ! NAME
    !
    !  Function mult_div
    !
    ! USAGE
    !
    !  real_dim4 = mult_div(alpha,array)
    !
    ! DESCRIPTION
    !
    !  This function computes numericaly (1+alpha*epsilon)*(A/epsilon+B)
    !  with A = a1 + i*a2 and B = b1 + i*b2. The returned result is put
    !  into an array t (rank 1, shape 4) where t(1) = a1, t(2) = a2,
    !  t(3) = b1+alpha*a1, t(4) = b2+alpha*a2. 
    !
    ! INPUTS
    !
    !  * alpha -- a real (type ki)
    !  * array -- a real (type ki) array of rank 1, shape 4
    !
    ! SIDE EFFECTS
    !
    !  No side effect
    !
    ! RETURN VALUE
    !
    !  a real (type ki) array of rank 1 and shape 4
    !
    ! EXAMPLE
    !
    !  resu = multipy_div(alpha,array)
    !  resu(1) = array(1)
    !  resu(2) = array(2)
    !  resu(3) = array(3) + alpha*array(1)
    !  resu(4) = array(4) + alpha*array(2)
    !
    !*****
    function mult_div(alpha,array)
      !
      real(ki), intent(in) :: alpha
      real(ki), intent(in), dimension(4) :: array
      real(ki), dimension(4) :: mult_div
      !
      mult_div = array
      mult_div(3) = mult_div(3) + alpha*array(1)
      mult_div(4) = mult_div(4) + alpha*array(2)
      !
    end function mult_div
    !
end module multiply_div
